/*
;
;---------------------------------------------------------------+
;       .__               __                          			
;______ |__|___________ _/  |_            ____  ______			
;\____ \|  \_  __ \__  \\   __\  ______  /  _ \/  ___/			
;|  |_> >  ||  | \// __ \|  |   /_____/ (  <_> )___ \ 			
;|   __/|__||__|  (____  /__|            \____/____  >			
;|__|                  \/                          \/ 			
;---------------------------------------------------------------+
;
;[1] Informations
;
; Last Modified: 	21. Oktober 2004
; Begin: 			15. Juni 2004
; Version: 			0.000
; Coder: 			z4ck
;
;
;[2] Tasks
;
;	Task									Done		Coder
;----------------------------------------------------------------
; -                  						[  0%]		z4ck
;----------------------------------------------------------------
; TOTAL										[  0%]		z4ck
;================================================================
*/

#include <x32.h>
#include <mm/mm.h>
#include <sys/port.h>


/***************************************/
//Structs
typedef struct tagMemBlock {
	uint32_t	*adress;
	uint32_t	size;
	
	struct tagMemBlock *next;

} __attribute__ ((packed)) MemBlock;

typedef struct tagMemList {
	MemBlock *head;
	MemBlock *tail;
} __attribute ((packed)) MemList;

/***************************************/
//Global Variables
uint32_t *PageDirectory;
uint32_t *PageTable;

//For the Heap
MemList FreeBlocks;
MemList UsedBlocks;
uint32_t HeapTail;
uint32_t HeapStart;


/**************************************************************************************************************************/
/**************************************************************************************************************************/
/**************************************************************************************************************************/
/**************************************************************************************************************************/
void SetupPaging()
{
	
	
	PageDirectory 	= (uint32_t *) 0x200000;
	PageTable 		= (uint32_t *) 0x201000;

	uint32_t i = 0;
	uint32_t j = 0;
	
	for (; i < 1024; i++)
	{
		PageDirectory[i] 	= 	(uint32_t)(PageTable + (4096 * i));
		PageDirectory[i] 	= 	PageDirectory[i] | PAGE_RW | PAGE_PRE;
	}
	
	MapMemory( 0, 0, (1024*1024), PAGE_PRE | PAGE_RW);
	
	//Kernel + BIOS + RM-VT ETC.. -> 0 - 0xFFFFF = 0x100 Pages
	MapMemory( 0, 							0, 							0x100, 	PAGE_RW | PAGE_PRE);
	
	//PageDirectory -> 0x100000 - 0x100FFF (1024 Entrys of 4 bytes) = 1 Page
	MapMemory( (uint32_t)PageDirectory, 	(uint32_t)PageDirectory, 	1, 		PAGE_RW | PAGE_PRE);
	
	//PageTable -> 0x101000 - 0x500FFF (1024 DirEntrys * 1024 TblEntry of 4 bytes) = 1024 Pages
	MapMemory( (uint32_t)PageTable, 		(uint32_t)PageTable, 		1024, 	PAGE_RW | PAGE_PRE);

	LoadPageDirectory((uint32_t)PageDirectory);
	
	FreeBlocks.head = NULL;
	FreeBlocks.tail = NULL;
	
	UsedBlocks.head = NULL;
	UsedBlocks.tail = NULL;
	
	HeapStart	= 0x601000;
	HeapTail 	= 0;
	
	
};

/**************************************************************************************************************************/
/**************************************************************************************************************************/
/**************************************************************************************************************************/
/**************************************************************************************************************************/
uint8_t LoadPageDirectory(uint32_t _pd)
{
	if (_pd & 0xFFF) return 0;
	WriteCr( _pd, 3);
	WriteCr( (ReadCr(0) | 0x80000000) , 0);
	
	return 1;
}


/**************************************************************************************************************************/
/**************************************************************************************************************************/
/**************************************************************************************************************************/
/**************************************************************************************************************************/
uint8_t SetPage ( uint32_t _vAdress, uint32_t _phyAdress, uint32_t _opt)
{

	_vAdress 	= _vAdress 		& 0xFFFFF000;
	_phyAdress 	= _phyAdress 	& 0xFFFFF000;
	
	uint32_t table 	= _vAdress >> 22;
	uint32_t entry  = (_vAdress << 10) >> 22;
	
	if (table < 1024 && entry < 1024)
	{
		uint32_t *PT = (uint32_t *)(PageDirectory[table] & 0xFFFFF000);
		PT[entry] = _phyAdress | _opt;
		return 1;
	}
	return 0;
};



/**************************************************************************************************************************/
/**************************************************************************************************************************/
/**************************************************************************************************************************/
/**************************************************************************************************************************/
uint8_t MapMemory ( uint32_t _vAdress, uint32_t _phyAdress, uint32_t _lenght, uint32_t _opt)
{
	while (_lenght > 0)
	{
		uint8_t state = SetPage(_vAdress, _phyAdress, _opt);
		_vAdress 	+= 0x1000;
		_phyAdress 	+= 0x1000;
		_lenght--;
		
		if (!state)
			return 0;
	}
	return 1;
};

/**************************************************************************************************************************/
/**************************************************************************************************************************/
/**************************************************************************************************************************/
/**************************************************************************************************************************/
int32_t GetEmptyPDE()
{
	uint32_t i;
	for (i = 0; i < 1024; i++)
	{
		if (PageDirectory[i] == 0) return i;
	
	}
	return -1;
};

/**************************************************************************************************************************/
/**************************************************************************************************************************/
/**************************************************************************************************************************/
/**************************************************************************************************************************/
int32_t GetEmptyPTE(uint32_t _pde)
{
	uint32_t *ptbl = (uint32_t *)(PageDirectory[_pde] & 0xFFFFF000);
	uint32_t i;
	for (i = 0; i < 1024; i++)
	{
		if (ptbl[i] == 0) return i;
	
	}
	return -1;
};


/**************************************************************************************************************************/
/**************************************************************************************************************************/
/**************************************************************************************************************************/
/**************************************************************************************************************************/
void *realloc(void *_adress, uint32_t _more)
{

	return (void *) 0;
};

/**************************************************************************************************************************/
/**************************************************************************************************************************/
/**************************************************************************************************************************/
/**************************************************************************************************************************/
void *malloc(uint32_t _size)
{
	/*********************************/
	/* Search for Blocks which have 
	   exact the same size as the 
	   requestet one.
	*/
	MemBlock *tmp 	= FreeBlocks.head;
	MemBlock *last	= NULL;
	
	if (tmp != NULL)	//If min. one Freeblock exists
	{	
		while (tmp->next != NULL && tmp->size != _size)
		{
			last = tmp;
			tmp = tmp->next;
		} 
		
		/*********************************/
		/* If not found search one which
		   is bigger wheter the requestet
		   one.
		*/
		if (tmp->size != _size)
		{
			tmp = FreeBlocks.head;
			last = NULL;
			while (tmp->next != NULL && tmp->size < _size)
			{
				last = tmp;
				tmp = tmp->next;
			} 
		}
		
		/*********************************/
		/* If a block is found which
		   is bigger or which has the
		   same size take it from the
		   FreeBlockList and put that
		   in the UsedBlockList.
		*/
		if (tmp->size >= _size)
		{
			if (last == NULL) //<-- If its the first entry
			{
				FreeBlocks.head = tmp->next;
				
				if (tmp->next == NULL)
				{
					FreeBlocks.tail = last;
					FreeBlocks.tail->next = NULL;
				}
			}
			else if (tmp->next == NULL) //<-- If its the last entry
			{
				FreeBlocks.tail = last;
				FreeBlocks.tail->next = NULL;
			}
			else						//<<- if its in the middle of the list
			{
				last->next = tmp->next;
			}
			
			if (UsedBlocks.tail != NULL) UsedBlocks.tail->next = tmp;
			if (UsedBlocks.head == NULL) UsedBlocks.head = tmp;
			
			UsedBlocks.tail = tmp;
			tmp->next = NULL;
			
			//make something if the free space is bigger wheter the requestet place...
			return (void *)tmp->adress;
		}
	
	}
	/*********************************/
	/* If no block was found make a new
	   one at the end of UsedBlocks.
	*/
	tmp = (MemBlock *) HeapStart + HeapTail;
	HeapTail += sizeof(MemBlock) + _size;
	
	if (UsedBlocks.tail != NULL) UsedBlocks.tail->next = tmp;
	if (UsedBlocks.head == NULL) UsedBlocks.head = tmp;
	
	UsedBlocks.tail = tmp;
	tmp->next = NULL;
	tmp->size = _size;
	tmp->adress = (uint32_t*)((uint32_t)tmp + sizeof(MemBlock));
	
	return (void *)tmp->adress;
};

/**************************************************************************************************************************/
/**************************************************************************************************************************/
/**************************************************************************************************************************/
/**************************************************************************************************************************/
int8_t free(void * _adress)
{
	MemBlock *tmp = UsedBlocks.head;
	MemBlock *last = NULL;
		
	while (tmp->next != NULL && tmp->adress != _adress)
	{
		last = tmp;
		tmp = tmp->next;
	} 
	
	if (tmp->adress != _adress)
	{
		return -1;
	}
	
	if (last == NULL)
	{
		UsedBlocks.head = tmp->next;
		if (tmp->next == NULL)
		{
			UsedBlocks.tail = last;
			UsedBlocks.tail->next = NULL;
		}
	}
	else if (tmp->next == NULL)
	{
		UsedBlocks.tail = last;
		UsedBlocks.tail->next = NULL;
	}
	else
	{
		last->next = tmp->next;
	}
	
	if (FreeBlocks.head == NULL) FreeBlocks.head = tmp;
	if (FreeBlocks.tail != NULL) FreeBlocks.tail->next = tmp;
	
	FreeBlocks.tail = tmp;
	tmp->next = NULL;
	
	return 1;
};

/**************************************************************************************************************************/
/**************************************************************************************************************************/
/**************************************************************************************************************************/
/**************************************************************************************************************************/
void *memcpy(void *_dest, void *_src, uint32_t _lenght)
{
	uint32_t i;
	
	unsigned char *from = (unsigned char*)_src;
	unsigned char *to   = (unsigned char*)_dest;
	
	for (i = 0; i < _lenght; i++)
		to[i] = from[i];
	
	return _dest;	
};


/**************************************************************************************************************************/
/**************************************************************************************************************************/
/**************************************************************************************************************************/
/**************************************************************************************************************************/
void *memset(void *_dest, unsigned char _char, unsigned long _lenght)
{
	uint32_t i = 0;
	unsigned char *to = (unsigned char *)_dest;
	
	while (i < _lenght)
	{
		to[i] = _char;
		i++;
	}
	
	return _dest;
};
