/*
;
;---------------------------------------------------------------+
;       .__               __                          			
;______ |__|___________ _/  |_            ____  ______			
;\____ \|  \_  __ \__  \\   __\  ______  /  _ \/  ___/			
;|  |_> >  ||  | \// __ \|  |   /_____/ (  <_> )___ \ 			
;|   __/|__||__|  (____  /__|            \____/____  >			
;|__|                  \/                          \/ 			
;---------------------------------------------------------------+
;
;[1] Informations
;
; Last Modified: 	02. Februar 2004
; Begin: 			02. Februar 2004
; Version: 			1.000
; Coder: 			z4ck
;
;
;[2] Tasks
;
;	Task									Done		Coder
;----------------------------------------------------------------
; -                  						[  0%]		z4ck
;----------------------------------------------------------------
; TOTAL										[  0%]		z4ck
;================================================================
*/

#include <task/multi.h>
#include <sys/error.h>
#include <mm/mm.h>
#include <task/list.h>
//#include <gui/uiClip.h>
//#include <driver/video.h>

/*------------GLOBAL VARS----------------*/
List 	*TaskList;
uint32_t TaskID;
uint8_t	 initialized = false;
TASK 	*CurrentTask;

////////////////////START REMOVE AFTER TESTING
void task1();
void task2();
//////////////////////////////////////////////////////////////END REMOVING


//This one setup the Scheduler. It must setup at least 2 tasks
// IDLE -> don't make anything
// And surely one task for the Kernel wich is executed first.
// The Kernel task is a litle bit special because he is runing till now
// and needn't a extra new stack 0.o
uint8_t SetupScheduler()
{
	TaskList = NewList();
	TaskID = 0;
	CurrentTask = NULL;
	
	SetupKernelTask();
	
	TaskAdd(task1, 		"task1", 	USER_TASK, 		NORMAL_PRIORITY);
	TaskAdd(task2, 		"task2", 	USER_TASK, 		NORMAL_PRIORITY);
//	TaskAdd(Task3, 		"task3", 	USER_TASK, 		NORMAL_PRIORITY);
	
	TaskAdd(IdleTask, 	"idle",  	KERNEL_TASK, 	SUPER_LOW_PRIORITY);
	
	
	
	initialized = true;
	
	return OK;
}

// The point of this method is to save the new stack top of the current task
// in the struct of the task. on which stack all registere were pushed.
// the select a new task and return his esp. back to the lowlevel routine
// asm pops the registers and we have loaded a new task. however, as simple 
// this one sounds, it brings problems with it...
// will be executed by isr.asm -> irq0
uint32_t Scheduler(uint32_t _oldEsp)
{
	//if the scheduler isn't right initialized return 
	//only the old esp...
	if (!initialized) 
		return _oldEsp;
	
	//this is only for the case that no Task is current running...
	if (CurrentTask != NULL)
		CurrentTask->Esp = _oldEsp; //Save the current ESP to the current task
	
	uint8_t		ratio  	= 2;
	uint32_t 	i 		= 0;
	uint32_t 	max 	= 0;
	uint32_t	weight  = 0;
	
	TASK * tmpTASK = (TASK *) ListGet(TaskList, i);
	TASK * maxTASK = NULL;
	
	while (tmpTASK != NULL)
	{	
		if (tmpTASK->State != TASK_SLEEPING) 
		{
			weight = ( tmpTASK->Priority * ratio ) + tmpTASK->Timetorun;
			if (weight > max)
			{
				max = weight;
				maxTASK = tmpTASK;
			}
				
			tmpTASK->Timetorun++;
		}	
		i++;
		tmpTASK = (TASK *) ListGet(TaskList, i);
	}
	
	maxTASK->Timetorun = 0;
	CurrentTask = maxTASK;
	
	return maxTASK->Esp;
	

}


// This is the most complicatet one. we must generate a new stack for the
// new task, also a new TAS struct, overgive the properties, add to the 
// tasklist and push register onto the stack...
uint32_t TaskAdd(void (*_entrypoint), uint8_t *_name, uint8_t _type, uint8_t _priority)
{
	TASK *tmpTASK = malloc(sizeof(TASK)); //Allocate  a new TASK struct
	
	//Seting various properties to the new TASK
	tmpTASK->ID = ++TaskID;
	tmpTASK->Priority = _priority;
	tmpTASK->Type = _type;
	tmpTASK->State = TASK_RUNNING;
	tmpTASK->Timetorun = 10;
	tmpTASK->box = msgsetup();
	strcpy(tmpTASK->Name, _name);
	
	//Allocating space for a 4kb stack
	uint32_t *stackbase = (uint32_t *) ((uint32_t)malloc(4096) + 4096);
	uint32_t *stackptr  = stackbase;
	
	//now pushing all registers to the stack
	*--stackptr = 0x0202; //EFLAGS				 #
	*--stackptr = 0x08;	  //CS					 |---> iret
	*--stackptr = (uint32_t) _entrypoint;//		 #
	
	*--stackptr = 0; //EBP					  	 #
	*--stackptr = 0; //ESP						 |
	*--stackptr = 0; //EDI						 |
	*--stackptr = 0; //ESI						 |----- pushad
	*--stackptr = 0; //EDX						 |
	*--stackptr = 0; //ECX						 |
	*--stackptr = 0; //EBX						 |
	*--stackptr = 0; //EAX						 #
	
	*--stackptr = 0x10; //DS
	*--stackptr = 0x10; //ES
	*--stackptr = 0x10; //FS
	*--stackptr = 0x10; //GS
	
	tmpTASK->Ss  = (uint32_t) stackbase;
	tmpTASK->Esp = (uint32_t) stackptr;
	
	ListAdd(TaskList, tmpTASK); // Add the Task to the TaskList
	
	return tmpTASK->ID;
}

uint8_t TaskWait(uint32_t _id);
uint8_t TaskSleep(uint32_t _id);
uint8_t TaskWake(uint32_t _id);
uint8_t TaskResume(uint32_t _id);
uint8_t TaskKill(uint32_t _id);
uint8_t TaskClone(uint32_t _id, uint32_t _id2);
uint8_t TaskSendName(uint32_t _id, uint32_t _id2);
uint8_t TaskSendId(uint32_t _id, uint32_t _id2);
uint8_t TaskRequestName(uint32_t _id, uint32_t _id2);
uint8_t TaskRequestId(uint32_t _id, uint32_t _id2);
uint32_t TaskBoxId()
{
 TASK *tmpTASK = TaskGetCurrent();
 return tmpTASK->box;
}
uint8_t TaskRemove(uint32_t _id)
{
	TASK *tmpTASK = TaskGetById(_id);
	
	if (tmpTASK == NULL)
		return ERR_NOT_FOUND;
		
	free(&tmpTASK->Ss);
	
	ListRemove(TaskList, tmpTASK);
	free(tmpTASK);

}

uint8_t TaskSetPriority(uint32_t _id, uint8_t _priority)
{
	TASK *tmpTASK = TaskGetById(_id);
	if (tmpTASK == NULL)
		return ERR_NOT_FOUND;
	
	tmpTASK->Priority = _priority;
	return OK;
}

uint8_t TaskSetState(uint32_t _id, uint8_t _state)
{
	TASK *tmpTASK = TaskGetById(_id);
	if (tmpTASK == NULL)
		return ERR_NOT_FOUND;
		
	tmpTASK->State = _state;
	return OK;
}

TASK * TaskGetCurrent()
{
	return CurrentTask;
}

TASK * TaskGetById(uint32_t _id)
{
	TASK *tmpTASK = NULL;
	
	uint32_t i = 0;
	
	do
	{
		tmpTASK = (TASK *) ListGet(TaskList, i);
		i++;
	} while (tmpTASK->ID != _id && i < TaskList->size); 
	
	return (TASK *) tmpTASK;
}

uint8_t SetupKernelTask()
{
	TASK *tmpTASK = malloc(sizeof(TASK)); //Allocate  a new TASK struct
	
	//Seting various properties to the new TASK
	tmpTASK->ID = ++TaskID;
	tmpTASK->Priority = NORMAL_PRIORITY;
	tmpTASK->Type = KERNEL_TASK;
	tmpTASK->State = TASK_RUNNING;
	tmpTASK->Timetorun = 0;
	
	strcpy(tmpTASK->Name, "kernel");
	
	ListAdd(TaskList, tmpTASK); // Add the Task to the TaskList
	
	CurrentTask = tmpTASK;
	return OK;
}

void IdleTask()
{
	/* This Task does hardly nothing except looping around          */
	/* This is mayby the most easy part of my os to understand.. :) */
	while(1);

}
